import re

# Nested Form data deserializer utility class
# --------------------------------------------

class UtilityMixin(object):

    _nested_re = re.compile(r'((.+)\[(.*)\]+)|(\[(.*)\]{2,})')
    _namespace_re = re.compile(r'^([\w]+)(?=\[)')
    _list_re = re.compile(r'\[([0-9]+)\]')
    _number_re = re.compile(r'[0-9]+')
    _dict_re = re.compile(r'\[(([^A-Za-z]*)([^0-9]+)([0-9]*))+\]')

    @staticmethod
    def strip_bracket(string=''):
        return string.replace('[', '').replace(']', '')

    @staticmethod
    def split(string=''):
        subkeys = [key + ']' for key in string.split(']') if key is not '']

        assert len(subkeys) > 0, ('Cannot split this key `%s`' % (string))

        return subkeys

    @staticmethod
    def is_empty_list(string=''):
        return string == '[]'

    @staticmethod
    def is_empty_dict(string=''):
        return string == '{}'

    def is_dict(self, string=''):
        return bool(self._dict_re.fullmatch(string))

    def is_list(self, string=''):
        return bool(self._list_re.fullmatch(string))

    def is_number(self, string=''):
        return bool(self._number_re.fullmatch(string))

    def is_nested_string(self, string=''):
        return bool(self._nested_re.fullmatch(string))

    def is_namespaced(self, string=''):
        return bool(self._namespace_re.match(string))

    def get_namespace(self, string=''):
        namespace = self._namespace_re.match(string)

        if namespace:
            return namespace.group(0)

        return None

    def strip_namespace(self, string=''):
        """
        Strip namespace from key if any 
        """
        namespace = self._namespace_re.match(string)

        if namespace:
            splited_str = string.split(namespace.group(1))

            assert len(splited_str) > 0, (
                'Cannot strip namespace from' 'this key `%s`' % (string)
            )

            string = ''.join(splited_str)

            return string

        return string

    def extract_index(self, string=''):
        number = self._number_re.search(string)

        if number:
            return int(number.group(0))

        return None

    def replace_specials(self, string):
        """
        Checks for and replaces special characters like null, booleans
        also changes numbers from string to integer
        """
        # the incoming value may not be a string sometimes
        if not isinstance(string, str):
            return string

        if string == 'null':
            return None
        elif string == 'true':
            return True
        elif string == 'false':
            return False
        elif self.is_number(string):
            return int(string)
        elif self.is_empty_list(string):
            return []
        elif self.is_empty_dict(string):
            return {}
        else:
            return string

    def get_container(self, key, index=0):
        """
        It return the appropiate container `[]`|`{}`
        based on the key provided
        """
        if self.is_nested_string(key):
            key = self.strip_namespace(key)
            sub_keys = self.split(key)

            if self.is_list(sub_keys[index]):
                return []
            elif self.is_dict(sub_keys[index]):
                return {}
            else:
                return None
        else:
            return {}

    def is_last(self, current_key, data, is_type_of='O'):
        """
        Checks if current key is the last in the data
        """
        keys = list(data.keys())
        current_index = keys.index(current_key)
        next_index = current_index + 1

        def object_type(is_obj):
            try:
                next_key = self.split(keys[next_index])[0]
                if is_obj(next_key):
                    is_last = False
                else:
                    is_last = True
            except IndexError:
                is_last = True

            return is_last

        def namespace():
            try:
                next_key = keys[next_index]
                if self.is_namespaced(next_key):
                    current_namespace = self.get_namespace(current_key)
                    next_namespace = self.get_namespace(next_key)

                    if current_namespace == next_namespace:
                        is_last = False
                    else:
                        is_last = True
                else:
                    is_last = True
            except IndexError:
                is_last = True

            return is_last

        def ordinary():
            try:
                if not self.is_namespaced(keys[next_index]):
                    next_k = self.split(keys[next_index])[0]

                    if not self.is_list(next_k) and not self.is_dict(next_k):
                        is_last = False
                    else:
                        is_last = True
                else:
                    is_last = True
            except IndexError:
                is_last = True

            return is_last

        if is_type_of == 'L':
            return object_type(self.is_list)
        elif is_type_of == 'N':
            return namespace()
        elif is_type_of == 'D':
            return object_type(self.is_dict)
        else:
            return ordinary()



from rest_framework.parsers import MultiPartParser, JSONParser

from .utils import NestedForms
from .settings import api_settings


class NestedMultiPartParser(MultiPartParser):
    """
    Parser for multipart form data that is nested and also
    it may include files
    """
    options = api_settings.OPTIONS

    def parse(self, stream, media_type=None, parser_context=None):
        parsed = super().parse(stream, media_type, parser_context)

        # files and data have to be merged into one
        if parsed.files:
            self._full_data = parsed.data.copy()
            self._full_data.update(parsed.files)
        else:
            self._full_data = parsed.data

        form = NestedForms(self._full_data, **self.options)

        if form.is_nested():
            return form.data

        return parsed


class NestedJSONParser(JSONParser):
    """
    Parser for JSON data that is nested
    """
    options = api_settings.OPTIONS

    def parse(self, stream, media_type=None, parser_context=None):
        parsed = super().parse(stream, media_type, parser_context)

        form = NestedForms(parsed, **self.options)

        if form.is_nested():
            return form.data

        return parsed



from collections.abc import Mapping

from .exceptions import ParseError
from .mixins import UtilityMixin


# Base serilizer for converting nested formdata to objects
# ---------------------------------------------------------

class BaseClass(UtilityMixin):

    def __init__(self, data, *args, **kwargs):
        self._initial_data = data
        self._allow_empty = kwargs.get('allow_empty', False)
        self._allow_blank = kwargs.get('allow_blank', True)

    def __process__(self):
        raise NotImplementedError('`__process__()` is not implemented')

    def __run__(self):
        if not hasattr(self, '_final_data'):
            self.__process__()

    @property
    def data(self):
        """
        Returns the final build
        """
        if not hasattr(self, '_final_data'):
            msg = '`.is_nested()` has to be called before accessing `.data`'
            raise AssertionError(msg)

        return self._final_data

    @property
    def validated_data(self):
        if not hasattr(self, '_validated_data'):
            msg = '`.is_nested()` has to be called before accessing `.validated_data`'
            raise AssertionError(msg)

        return self._validated_data

    def is_nested(self, raise_exception=False):
        """
        Checks if the initial_data map is a nested object
        """

        ###### Check if initial_data is a MultiValueDIct ############
        ###### Convert it to a dict object ##########################

        if hasattr(self._initial_data, 'getlist'):
            raw_data = {}

            for key, value in dict(self._initial_data).items():
                if len(value) > 1:
                    raw_data[key] = value
                else:
                    raw_data[key] = value[0]

            self._initial_data = raw_data

        #############################################################

        is_mapping = isinstance(self._initial_data, Mapping)
        conditions = [is_mapping]

        #############################################################

        if not is_mapping and raise_exception:
            raise ValueError('`data` is not a map type')

        #############################################################

        matched_keys = []

        for key in self._initial_data.keys():
            if self.is_nested_string(key):
                matched_keys.append(True)
                break
            else:
                matched_keys.append(False)

        conditions += [any(matched_keys)]

        #############################################################

        if not any(matched_keys) and raise_exception:
            raise ValueError('`data` is not a nested type')

        #############################################################
        
        self._validated_data = self._initial_data
        self.__run__()

        return all(conditions)

    def clean_value(self, value, default=None):
        """
        Replace empty list, dict and empty string with `None`
        """
        if value == '' and not self._allow_blank:
            return default
        elif isinstance(value, list) and len(value) == 0 and not self._allow_empty:
            return default
        elif isinstance(value, dict) and len(value) == 0 and not self._allow_empty:
            return default

        return value


# serilizes a nested form data to object
# ---------------------------------------

class NestedForms(BaseClass):
    """
    decode nested forms into python object
    """

    def __process__(self):
        """
        Initiates the conversion process and packages the final data
        """
        data_list, final_build, top_wrapper = self._group_data(), {}, []

        for data in data_list:
            key = list(data.keys())[0]
            value = list(data.values())[0]

            root_tree = self.decode(value)

            if not bool(key):
                if isinstance(root_tree, dict):
                    final_build.update(root_tree)
                elif isinstance(root_tree, list):
                    if final_build:
                        final_build[''] = root_tree
                    else:
                        top_wrapper.append(root_tree)
            else:
                final_build.setdefault(key, root_tree)

        if final_build:
            top_wrapper.append(final_build)

        if len(top_wrapper) == 1:
            self._final_data = top_wrapper[0]
        elif len(top_wrapper) > 1:
            self._final_data = top_wrapper
        else:
            raise ParseError('unexpected empty container')

    def _group_data(self):
        """
        It groups data structures of the same kind and namespaces together
        """
        temp, container, data = {}, [], self.validated_data

        for key, value in data.items():
            if self.is_namespaced(key):
                if self.is_last(key, data, 'N'):
                    temp.setdefault(key, value)
                    container.append({self.get_namespace(key): temp})
                    temp = {}
                else:
                    temp.setdefault(key, value)
            elif self.is_list(self.split(key)[0]):
                if self.is_last(key, data, 'L'):
                    temp.setdefault(key, value)
                    container.append({'': temp})
                    temp = {}
                else:
                    temp.setdefault(key, value)
            elif self.is_dict(self.split(key)[0]):
                if self.is_last(key, data, 'D'):
                    temp.setdefault(key, value)
                    container.append({'': temp})
                    temp = {}
                else:
                    temp.setdefault(key, value)
            else:
                if self.is_last(key, data):
                    temp.setdefault(key, value)
                    container.append({'': temp})
                    temp = {}
                else:
                    temp.setdefault(key, value)

        return container

    def decode(self, validated_data):
        """
        Convert validated_data to an object containing primitives
        """
        self._root_tree = None

        ############# initialize the root_tree ###################

        if self._root_tree is None:
            key = list(validated_data.keys())[0]
            self._root_tree = self.get_container(key)

        #########################################################

        for key, value in validated_data.items():
            if self.is_nested_string(key):
                key = self.strip_namespace(key)
                value = self.replace_specials(value)

                value = self.clean_value(value)
                self.generate_context(key, value, self._root_tree)
            else:
                # the root tree is automatically a dict
                self._root_tree.setdefault(key, value)

        return self._root_tree

    def build(self, root, context, depth=0):
        """
        Insert and updates the root tree according to the context passed as
        argument
        """

        def is_dict(v): return isinstance(v, dict)
        def is_list(v): return isinstance(v, list)

        ######################################################

        def set_root():
            if is_list(root):
                # check the difference between index of the last item in
                # the list and the current index to be added
                diff = abs(len(root) - context['index'])

                if diff > 1000:
                    raise ParseError('too many consecutive empty arrays !')
                elif diff > 1:
                    # if it is sparse
                    # fill gaps with the `None`
                    for _ in range(diff):
                        root.append(None)
                    # append the default value
                    root.append(context['value'])
                elif context['value'] and is_list(context['value']):
                    # if context value is not empty and its a list, transfer
                    # value into the root
                    root.extend(context['value'])
                else:
                    # just append like default
                    root.append(context['value'])
            else:
                root[context['index']] = context['value']

        ######################################################

        if context['depth'] == depth:
            try:
                # check if the root of interest is empty or exist
                value = root[context['index']]
                if value:
                    if is_list(value) and context['value']:
                        value.append(context['value'])
                    elif is_dict(value) and is_dict(context['value']):
                        for key in context['value'].keys():
                            # if the key already exist do nothing
                            if key in value:
                                continue
                            else:
                                # transfer all the value to the root or value
                                value.update(context['value'])
                    else:
                        pass
                else:
                    set_root()
            except (KeyError, IndexError):
                set_root()
        else:
            # if not depth of interest unpack the root object with keys
            # provided by context['keys']
            key = context['keys'][depth]
            self.build(root[key], context, depth=depth + 1)

    def generate_context(self, key, value, root_tree):
        """
        Generates a context for every key
        """
        sub_keys = self.split(key)
        index_keys = []

        def get_index(key):
            if self.is_dict(key):
                return self.strip_bracket(key)
            elif self.is_empty_list(key):
                # an empty list always has it index as '0'
                # unless the key is repeated, then it will
                # have an array of values attached to the same key
                return 0

            return self.extract_index(key)

        def get_value(index):
            if index < len(sub_keys):
                if self.is_dict(sub_keys[index]):
                    # at this time the value of the dict is unknown
                    # so `None` is used
                    return {self.strip_bracket(sub_keys[index]): None}
                else:
                    return []

            return value

        def get_index_keys(index):
            if index >= 0:
                index_keys.append(get_index(sub_keys[index]))
                return index_keys

            return index_keys

        #########################################################

        for i, sub_key in enumerate(sub_keys):

            next_i = i + 1
            prev_i = i - 1

            context = {
                'depth': i,
                'index': get_index(sub_key),
                'value': get_value(next_i),
                'keys': get_index_keys(prev_i)
            }

            self.build(root_tree, context)



__title__ = 'drf_nested_formdata'
__version__ = '0.2.1'
__author__ = 'Duke Effiom'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020'

# Version synonym
from .exceptions import ParseError
from .parsers import NestedMultiPartParser, NestedJSONParser
from .mixins import UtilityMixin
from .utils import NestedForms
VERSION = __version__



from django.conf import settings
from rest_framework.settings import APISettings


USER_SETTINGS = getattr(settings, 'NESTED_FORM_PARSER', {})

DEFAULTS = {
    'OPTIONS': {
        'allow_empty': False,
        'allow_blank': True
    }
}

api_settings = APISettings(USER_SETTINGS, DEFAULTS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('drf_nested_forms')

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='drf_nested_forms',
    version=version,
    url='http://github.com/emperorDuke/nested_formdata',
    download_url='http://github.com/emperorDuke/nested_formdata/archive/v0.2.1.tar.gz',
    license='MIT',
    description='A library that parses nested json or form data to python object',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['drf', 'nested form', 'html_forms' 'drf_nested_forms', 'restframework', 'nested json'],
    author='Duke Effiom',
    author_email='effiomduke@gmail.com',
    packages=['drf_nested_forms'],
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: MIT License',
    ], 
    test_suite='tests',
    tests_require=[
        'djangorestframework',
        'django'
    ]
)
